/*
 * lcd1602.c
 *
 *  Created on: 14.6.2011
 *      Author: emlin
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/kernel.h>

#include <asm/uaccess.h>

#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/delay.h>

// -----------------------------------------------------------------------------
#define DEV_NAME					"lcd1602"
#define DEV_COUNT					1

#define PAR_PORT_BASE				0x378
#define PAR_PORT_SIZE				3
// ofset od zacatku bloku registru paralelniho portu
#define PAR_PORT_REG_DATA			(0x00)
#define PAR_PORT_REG_STAT			(0x01)
#define PAR_PORT_REG_CTRL			(0x02)

// -----------------------------------------------------------------------------
// Hodnoty signalu /SEL, /AF a /STR jsou HW invertovane
// ridi signal RS, ktery prepina mezi zapisem dat(znaku)/instrukce
#define LCD_WRITE_CMD				0x09	// /SEL = 1, /INT = 0, /AF = 0, /STR = 1
#define	LCD_WRITE_CHAR				0x08	// /SEL = 1, /INT = 0, /AF = 0, /STR = 0
// ridi signal E, ktery slouzi pro de/aktivaci displeje
// na sestupne hrane signalu E se provede zapis dat do LCD displeje, ktery
// pak hned odpoji od sbernice
#define	LCD_RAISE_E					0x09	// /SEL = 1, /INT = 0, /AF = 0, /STR = 1
#define LCD_FALLING_E				0x0B	// /SEL = 1, /INT = 0, /AF = 1, /STR = 1

// -----------------------------------------------------------------------------
// Lcd1602Cmd_Init: 1, 8bit, 2 radky, znaky 5x7 bodu, 0, 0
#define LCD_CMD_SETUP				0x38
// Lcd1602Cmd_SetInputMode: 1, I/D = 1 (kurzoru dopredu), S = 0 (neposouvat displej)
#define LCD_CMD_SET_INPUT_MODE		0x06
// Lcd1602Cmd_Clear: 1 (smazat displej, kurzor na zacatek)
#define LCD_CMD_CLEAR				0x01
// Lcd1602Cmd_DisplayOn: 1, D = 1 (displej zapnut), C = 0 (kurzor vypnut),
//                   B = 0 (blikani kurzoru vypnuto)
#define LCD_CMD_DISPLAY_ON			0x0C
// Lcd1602Cmd_DisplayOff: 1, D = 0, C = 0, B = 0
#define	LCD_CMD_DISPLAY_OFF			0x08
// Lcd1602Cmd_SetAddr1R: DDRAM adresa 0x00 + horni bit na 1
#define	LCD_CMD_SET_ADDR_1R			0x80
// Lcd1602Cmd_SetAddr2R: DDRAM adresa 0x40 + horni bit na 1
#define	LCD_CMD_SET_ADDR_2R			0xC0

// -----------------------------------------------------------------------------
#define BUFFER_SIZE					32		// Bytes
#define MAX_ROWS					2
#define MAX_CHARS_PER_ROW			16

// -----------------------------------------------------------------------------
enum
{
	ModuleExit = 0,

	Failed_ChrdevRegion,
	Failed_CdevAlloc,
	Failed_CdevAdd,
	Failed_SysClass,
	Failed_SysDev,
	Failed_ReqRegion,
	Failed_IoMap,
};

// -----------------------------------------------------------------------------
// parametry ovladace
struct lcd1602_t
{
	// major-minor par
	dev_t				dev_num;
	// ukazatel na strukturu cdev reprezentujici znakove zarizeni
	struct cdev *		cdev_ptr;
	// ukazatel na tridu zarizeni v adresari /sys/class
	struct class *		sys_class_ptr;
	struct device * 	sys_dev_ptr;
	// hlida pocet volani operace open()
	atomic_t			open_cnt;
	// zacatek bloku registru paralelniho portu namapovaneho jako I/O pamet
	void __iomem *		port_base;
	// buffer znaku
	unsigned char		buffer[BUFFER_SIZE];
};

static struct lcd1602_t lcd =
{
	.dev_num = MKDEV(0, 0),
	.cdev_ptr = NULL,
	.sys_class_ptr = NULL,
	.sys_dev_ptr = NULL,
	.open_cnt = ATOMIC_INIT(0),
	.port_base = NULL,
};

// -----------------------------------------------------------------------------
static inline void wait_short(void)
{
	udelay(50); // usecs
}

static inline void wait_long(void)
{
	msleep(2); // msecs
}

static inline void wait_powerup(void)
{
	msleep(30); // msecs
}

// -----------------------------------------------------------------------------
static inline void write_raw(unsigned char byte, unsigned char writeCtrl)
{
	// nastavime signal RS a signal E na log.1
	iowrite8(LCD_RAISE_E & writeCtrl, (lcd.port_base + PAR_PORT_REG_CTRL));
	wmb();
	// na D0-D7 vystavime data
	iowrite8((byte), (lcd.port_base + PAR_PORT_REG_DATA));
	wmb();
	// shodime signal E do log.0, na sestupne hrane se zapisi data do lcd
	iowrite8(LCD_FALLING_E,  (lcd.port_base + PAR_PORT_REG_CTRL));
	wmb();
}

// -----------------------------------------------------------------------------
static void write_data(unsigned char data)
{
	// zapiseme data
	write_raw(data, LCD_WRITE_CHAR);
	// pockame az si to LCD prebere - neuime cist BUSY flag
	wait_short();
}

// -----------------------------------------------------------------------------
static void write_cmd(unsigned char cmd)
{
	// zapiseme instrukci
	write_raw(cmd, LCD_WRITE_CMD);
	// pockame si definovanu dobu podle typu instrukce
	switch ( cmd )
	{
		case LCD_CMD_CLEAR:
			wait_long();
			break;

		default:
			wait_short();
			break;
	}
}

// -----------------------------------------------------------------------------
// Funkce: Zapise data na displej s moznosti urceni polohy pomoci escape znaku:
// 			 '\t' -- posun o jeden znak doleva
// 			 '\r' -- skok na zacatek dalsiho radku
//
//		   Priklad:
//			echo -en "\t\t\tDobry den,\r\t\t\t\tmistre!" >/dev/lcd1602
//
//          Volba -e zapina interpretaci escape znaku (\t a \r jsou zmeneny na
//					byte 0x09 a 0x0D).
//          Volba  -n pak vypina pridavani znaku konce radku '\n'.
// -----------------------------------------------------------------------------
static void write_buf(void * data_ptr, size_t data_size)
{
	int i, charcnt = 0, set_addr = 0;
	unsigned char * buffer = (unsigned char *) data_ptr;

	// zacneme vzdy na prvnim znaku prvniho radku
	// pri zapisu vice znaku nez co se vejde na LCD, zajisti reset na pozici
	// prvni radek prvni znak
	write_cmd(LCD_CMD_SET_ADDR_1R);

	for ( i = 0; i < data_size; i++ )
	{
		// posun se o jeden znak doleva
		if ( buffer[i] == '\t' )
		{
			++charcnt;
			set_addr = 1;
		}
		// skoc na novy radek - dopln chybejici pocet znaku do konce radku
		else if ( buffer[i] == '\r' )
		{
			charcnt += MAX_CHARS_PER_ROW - (charcnt % MAX_CHARS_PER_ROW);
			set_addr = 1;
		}
		// vytiskni znak, nejprve nastav spravnou adresu v LCD
		else
		{
			// posun pozici znaku
			if ( set_addr )
			{
				unsigned int rows = charcnt / MAX_CHARS_PER_ROW;
				unsigned int cols = charcnt % MAX_CHARS_PER_ROW;
				unsigned int base = ( ((rows % MAX_ROWS) == 0) ?
												LCD_CMD_SET_ADDR_1R :
												LCD_CMD_SET_ADDR_2R );
				// zapis novou adresu do LCD
				write_cmd(base + cols);
				set_addr = 0;
			}

			// tisk znaku na LCD displeji
			write_data(buffer[i]);
			++charcnt;

			// po kazdem 16. znaku zmen adresu pro novy znak
			if ( !(charcnt % MAX_CHARS_PER_ROW) )
			{
				set_addr = 1;
			}
		}
	}
}

// -----------------------------------------------------------------------------
static void __init init_lcd(void)
{
	// cekaci doba po zapnuti zapajeni
	wait_powerup();

	// nastavit zakladni parametry: 8bit, 2 radky, znaky 5x7 bodu
	write_cmd(LCD_CMD_SETUP);

	// zapnout displej
	write_cmd(LCD_CMD_DISPLAY_ON);

	// smazat displej - mohou tam byt nahodna data
	write_cmd(LCD_CMD_CLEAR);

	// nastavit posun kurzoru (ano) a posun displeje (ne)
	write_cmd(LCD_CMD_SET_INPUT_MODE);
}

// -----------------------------------------------------------------------------
static int lcd1602_open(struct inode * inode_ptr, struct file * file_ptr)
{
	int result = 0, tmp_open_cnt = 0;

	// soubor zarizeni muze byt otevren jen jednim programem soucasne
	// atomicky inkrementuje promenou o 1 a vrati PUVODNI hodnotu
	tmp_open_cnt = atomic_inc_and_test(&lcd.open_cnt);
	if ( tmp_open_cnt > 1 )
	{
		pr_err("lcd1602_open(): already opened!\n");
		atomic_dec(&lcd.open_cnt);
		result = -EBUSY;
		goto End;
	}

	// modul nepodporuje systemove volani lseek()
	result = nonseekable_open(inode_ptr, file_ptr);
	if ( result < 0 )
	{
		goto End;
	}

	// smazeme displej
	write_cmd(LCD_CMD_CLEAR);

End:
	pr_devel("lcd1602_open(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static int lcd1602_release(struct inode * inode_ptr, struct file * file_ptr)
{
	// povolime nove otevreni
	atomic_dec(&lcd.open_cnt);

	pr_devel("lcd1602_release(): done\n");

	return 0;
}

// -----------------------------------------------------------------------------
static ssize_t lcd1602_read(struct file * file_ptr,
				   		 	char __user * data_ptr, size_t data_size,
				   		 	loff_t * pos_ptr)
{
	pr_devel("lcd1602_read(): failed\n");

	return -EINVAL; // nelze cist;
}

// -----------------------------------------------------------------------------
static ssize_t lcd1602_write(struct file * file_ptr,
				    		 const char __user * data_ptr, size_t data_size,
				    		 loff_t * pos_ptr)
{
	ssize_t result = 0;

	if ( data_size > 0 )
	{
		const size_t size_to_copy = min(sizeof(lcd.buffer), data_size);
		long res;

		// zkopirujeme znaky z uzivatelskeho prostoru, ale jen tolik znaku,
		// kolik se jich vejde na displej
		// (do tohoto limitu se zapocitavaji i ridici znaky)
		res = copy_from_user(&lcd.buffer, data_ptr, size_to_copy);
		if ( res < 0 )
		{
			result = -EFAULT;
			goto End;

		}

		// zapiseme data na displej
		write_buf(lcd.buffer, size_to_copy);
		result = size_to_copy;
	}

	pr_devel("lcd1602_write(): %d\n", result);

End:
	return result;
}

// -----------------------------------------------------------------------------
// virtualni tabulka - tabulka operaci podporovanych ovladacem
static struct file_operations lcd1602_file_ops =
{
	.owner = THIS_MODULE,
	.open = &lcd1602_open,
	.release = &lcd1602_release,
	.read = &lcd1602_read,
	.write = &lcd1602_write,
	// modul nepodporuje metodu llseek
	.llseek = &no_llseek,
};

// -----------------------------------------------------------------------------
static void cleanup(int error)
{
	switch ( error )
	{
		// kompletni uklid pri exitu
		case ModuleExit:
			// zrusime adresar modulu v /sys
			device_destroy(lcd.sys_class_ptr, lcd.dev_num);

		case Failed_SysDev:
			// zrusime tridu
			class_destroy(lcd.sys_class_ptr);

		case Failed_SysClass:
			// vyjmeme modul z jadra
			// pozn.: cdev_del() se vola jen po uspesnem cdev_add()
			cdev_del(lcd.cdev_ptr);

		case Failed_CdevAdd:
			// zrusime mapovani portu
			ioport_unmap(lcd.port_base);

		case Failed_IoMap:
			// zrusime rezervaci I/O portu
			release_region(PAR_PORT_BASE, PAR_PORT_SIZE);

		case Failed_ReqRegion:
			// pozn.: neexistuje opak fce cdev_alloc(), uvolneni pameti resi
			// jadro samo
		case Failed_CdevAlloc:
			// uvolnime nami zabrany major-minor par
			unregister_chrdev_region(lcd.dev_num, DEV_COUNT);

		case Failed_ChrdevRegion:
			// jeste nic se nealokovalo, takze neni co vracet
			break;

		default:
			break;
	}
}

// -----------------------------------------------------------------------------
static int __init lcd1602_init(void)
{
	int result = 0;
	struct resource * region_ptr;

	// registrace k major-minor paru
	result = alloc_chrdev_region(&lcd.dev_num, MINOR(lcd.dev_num),
								 DEV_COUNT, DEV_NAME);
	if ( result < 0 )
	{
		cleanup(Failed_ChrdevRegion);
		goto End;
	}

	// alokace pameti pro strukturu cdev
	lcd.cdev_ptr = cdev_alloc();
	if ( lcd.cdev_ptr == NULL )
	{
		result = -ENOMEM;
		cleanup(Failed_CdevAlloc);
		goto End;
	}

	// nastaveni vlastnika a seznamu souborovych operaci
	lcd.cdev_ptr->owner = THIS_MODULE;
	lcd.cdev_ptr->ops = &lcd1602_file_ops;

	// rezervujeme si paralelni port
	region_ptr = request_region(PAR_PORT_BASE, PAR_PORT_SIZE, DEV_NAME);
	if ( region_ptr == NULL )
	{
		pr_err("lcd1602_init(): parallel port already in use!\n");
		cleanup(Failed_ReqRegion);
		result = -EBUSY;
		goto End;
	}

	// premapujem I/O port na I/O pamet
	lcd.port_base = ioport_map(PAR_PORT_BASE, PAR_PORT_SIZE);
	if ( lcd.port_base == NULL )
	{
		cleanup(Failed_IoMap);
		result = -ENOMEM;
		goto End;
	}

	// inicializujeme LCD displej pro prvni pouziti
	// nastaveni rezimu prenosu, vstupniho rezimu apod.
	init_lcd();

	//  --- pridani ovladace do jadra ---
	result = cdev_add(lcd.cdev_ptr, lcd.dev_num, DEV_COUNT);
	if ( result < 0 )
	{
		cleanup(Failed_CdevAdd);
		goto End;
	}

	// vytvorime si tridu v /sys/class
	lcd.sys_class_ptr = class_create(THIS_MODULE, DEV_NAME);
	if ( IS_ERR(lcd.sys_class_ptr) )
	{
		cleanup(Failed_SysClass);
		result = PTR_ERR(lcd.sys_class_ptr);
		goto End;
	}

	// pridame nas modul do /sys/class/lcd1602
	lcd.sys_dev_ptr = device_create(lcd.sys_class_ptr,
									NULL, lcd.dev_num, NULL, DEV_NAME);
	if ( IS_ERR(lcd.sys_dev_ptr) )
	{
		cleanup(Failed_SysDev);
		result = PTR_ERR(lcd.sys_dev_ptr);
		goto End;
	}

End:
	pr_devel("lcd1602_init(): %d\n", result);

	return result;
}

// -----------------------------------------------------------------------------
static void __exit lcd1602_exit(void)
{
	cleanup(ModuleExit);

	pr_devel("lcd1602_exit(): done\n");
}

// -----------------------------------------------------------------------------
module_init(lcd1602_init);
module_exit(lcd1602_exit);

MODULE_AUTHOR("mBoss <marek.laznicka@elvoris.cz>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("2x16 LCD driver");
MODULE_VERSION("0.0.1");
